// MIDI Piano, is based on PBPiano from Borje Hagsten, June, 2016.
// C++ version (VS Community 2015) by Patrice Terrier, June, 2016.
// www.zapsolution.com

#include <windows.h>
#include <Commctrl.h>

#pragma warning(disable: 4996) // remove Unsafe notifications
#pragma warning(disable: 4312) // remove warning C4312: 'type cast': conversion from 'long' to 'HMENU' of greater size

#include "Tools.h"
#include "Piano.h"

//#include "GDImage.h"
//#pragma comment(lib, "..\\GDImage64.lib")
//#pragma comment(lib, "..\\WinLIFT64.lib")

// Enable tooltip
#pragma comment(linker,"\"/manifestdependency:type='win32' \
name='Microsoft.Windows.Common-Controls' version='6.0.0.0' \
processorArchitecture='*' publicKeyToken='6595b64144ccf1df' language='*'\"")

#define IDI_ICON1     101

#define IDC_LABEL1    101  // octave info
#define IDC_LABEL2    103  // < "button"
#define IDC_LABEL3    104  // > "button"
#define IDC_LABEL4    105  // F1 - help text
#define IDC_FRAME1    111  // Options
#define IDC_FRAME2    112  // Patch/Instrument
#define IDC_FRAME3    113  // Effects
#define IDC_FRAME4    114  // Pitch bend
#define IDC_CHECKBOX1 133  // Always on top
#define IDC_CHECKBOX2 131  // Show Key text
#define IDC_CHECKBOX3 132  // Show Note text
#define IDC_CHECKBOX4 134  // Sustain
#define IDC_LISTBOX1  141  // Instrument listbox
#define IDC_COMBOBOX1 151  // Channels ComboBox
#define IDC_COMBOBOX2 152  // Midi devices ComboBox
#define IDC_COMBOBOX3 153  // Instrument categories ComboBox
#define IDC_KEYBOARD  160  // Graphic control - piano keys
#define IDC_TRACKBAR1 181  // Volume
#define IDC_TRACKBAR2 182  // Balance
#define IDC_TRACKBAR3 183  // Vibrato
#define IDC_TRACKBAR4 184  // Pitch bend
#define IDC_PIANO     200  // PBPiano custom control

#define ClientW       634
#define ClientH       380

struct PROP {
    HFONT       usefont;
    HINSTANCE   hinstance;
    HWND        hMain;
    HWND        hPiano;
    long        minTrackSizeW;
    long        minTrackSizeH;
};

PROP gP;

void SetCtrlFont(IN HWND hWnd, IN HFONT usefont) {
    if (IsWindow(hWnd)) {
        if (usefont) { SendMessage(hWnd, WM_SETFONT, (WPARAM) usefont, 0); }
    }
}

LRESULT CALLBACK WndProc(IN HWND hWnd, IN UINT uMsg, IN WPARAM wParam, IN LPARAM lParam) {
    long c = 0, d = 0;
    WCHAR zText[1024] = { 0 };
    static BYTE F1Help;

    switch (uMsg) {
    case WM_GETDLGCODE: // Ensure the control processes all keys by itself
        return DLGC_WANTALLKEYS;

    case WM_KEYDOWN:    // keyboard keys
    case WM_KEYUP:
        SendMessage(gP.hPiano, uMsg, wParam, lParam);
        return 0;

    case WM_COMMAND:
        switch LOWORD(wParam) {
        case IDCANCEL:
            if (HIWORD(wParam) == BN_CLICKED) { 
                SendMessage(hWnd, WM_DESTROY, wParam, lParam);
                return 0;
            }
            break;

        case IDC_LABEL2:  // Arrow left < - octave change down
        case IDC_LABEL3:
            if ((HIWORD(wParam) == STN_CLICKED) || (HIWORD(wParam) == STN_DBLCLK)) {
                c = (long) SendMessage(gP.hPiano, PM_GETOCTAVE, 0, 0);
                d = (long) SendMessage(gP.hPiano, PM_GETKEYCOUNT, 0, 0);
                if (LOWORD(wParam) == IDC_LABEL2) {
                    c = max(0, c - 1);
                } else {
                    c = min(11 - d / 12, c + 1);
                }
                SendMessage(gP.hPiano, PM_SETOCTAVE, c, 1);
                wcscpy_s(zText, L"Octave "); wcsncat_s(zText, STRL(c), _TRUNCATE); wcsncat_s(zText, L" to ", _TRUNCATE); 
                wcsncat_s(zText, STRL(c + d / 12 - 1), _TRUNCATE); wcsncat_s(zText, L"\n(< or > to change)", _TRUNCATE);
                SetWindowText(GetDlgItem(hWnd, IDC_LABEL1), zText);
            }
            break;

        case IDC_CHECKBOX1:  // Topmost or not on screen
            if (HIWORD(wParam) == BN_CLICKED) {
                if (SendMessage((HWND) lParam, BM_GETCHECK, 0, 0)) {
                    SetWindowPos(hWnd, HWND_TOPMOST, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE);
                } else {
                    SetWindowPos(hWnd, HWND_NOTOPMOST, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE);
                }
                SetFocus(gP.hPiano);
            }
            break;

        case IDC_CHECKBOX2: // Show key text? 
            if (HIWORD(wParam) == BN_CLICKED) {
                c = (long) SendMessage((HWND) lParam, BM_GETCHECK, 0, 0);
                SendMessage(gP.hPiano, PM_SETKEYTEXT, c, 1);
            }
            break;

        case IDC_CHECKBOX3:  // Show note text? 
            if (HIWORD(wParam) == BN_CLICKED) {
                c = (long) SendMessage((HWND) lParam, BM_GETCHECK, 0, 0);
                SendMessage(gP.hPiano, PM_SETNOTETEXT, c, 1);
            }
            break;

        case IDC_CHECKBOX4:  // Sustain on/off
            if (HIWORD(wParam) == BN_CLICKED) {
                c = (long) SendMessage((HWND) lParam, BM_GETCHECK, 0, 0);
                SendMessage(gP.hPiano, PM_SETSUSTAIN, c, 1);
            }
            break;

        case IDC_COMBOBOX1:  // device
            if (HIWORD(wParam) == CBN_SELCHANGE) {
                c = (long) SendMessage((HWND) lParam, CB_GETCURSEL, 0, 0);
                SendMessage(gP.hPiano, PM_SETDEVICE, c, 0);
            } else if (HIWORD(wParam) == CBN_CLOSEUP) {
                SetFocus(gP.hPiano);
            }
            break;

        case IDC_COMBOBOX2:  // Channel
            if (HIWORD(wParam) == CBN_SELCHANGE) {
                c = (long) SendMessage((HWND) lParam, CB_GETCURSEL, 0, 0);
                SendMessage(gP.hPiano, PM_SETCHANNEL, c, 0);
            } else if (HIWORD(wParam) == CBN_CLOSEUP) {
                SetFocus(gP.hPiano);
            }
            break;

        case IDC_COMBOBOX3:  // Patches
            if (HIWORD(wParam) == CBN_SELCHANGE) {
                c = (long) SendMessage((HWND) lParam, CB_GETCURSEL, 0, 0);
                SendMessage(gP.hPiano, PM_LISTINSTRUMENTS, IDC_LISTBOX1, c);
                SendMessage(GetDlgItem(hWnd, IDC_LISTBOX1), LB_SETCURSEL, 0, 0);
                c = (long) SendMessage(GetDlgItem(hWnd, IDC_LISTBOX1), LB_GETITEMDATA, 0, 0);
                SendMessage(gP.hPiano, PM_SETINSTRUMENT, c, 1);
            } else if (HIWORD(wParam) == CBN_CLOSEUP) {
                SetFocus(gP.hPiano);
            }
            break;

        case IDC_LISTBOX1:  // Instruments
            if (HIWORD(wParam) == LBN_SELCHANGE) {
                c = (long) SendMessage((HWND) lParam, LB_GETCURSEL, 0, 0);
                c = (long) SendMessage((HWND) lParam, LB_GETITEMDATA, c, 0);
                SendMessage(gP.hPiano, PM_SETINSTRUMENT, c, 1);
            }
            break; 

        case IDC_PIANO:
            switch HIWORD(wParam) {
            case PN_PLAYNOTE:   // when a note is played, notification is sent here
                break;
            case PN_STOPNOTE:   // when a note is stopped, notification is sent here
                break;
            case PN_PITCHBEND:  // on up/down arrow key, pitch bend notification is sent here
                SendMessage(GetDlgItem(hWnd, IDC_TRACKBAR4), TBM_SETPOS, TRUE, lParam);
                break;
            case PN_OCTAVE:     // on left/right arrow key, octave change notification is sent here
                d = (long) SendMessage(gP.hPiano, PM_GETKEYCOUNT, 0, 0);
                wcscpy_s(zText, L"Octave "); wcsncat_s(zText, STRL((long) lParam), _TRUNCATE); wcsncat_s(zText, L" to ", _TRUNCATE); 
                wcsncat_s(zText, STRL(c + d / 12 - 1), _TRUNCATE); wcsncat_s(zText, L"\n(< or > to change)", _TRUNCATE);
                SetWindowText(GetDlgItem(hWnd, IDC_LABEL1), zText);
            }
            break;
        }
        break;

    case WM_CTLCOLOREDIT:
    case WM_CTLCOLORLISTBOX:
        SetBkColor((HDC) wParam, GetSysColor(COLOR_INFOBK));
        SetTextColor((HDC) wParam, GetSysColor(COLOR_INFOTEXT));
        return (LRESULT) GetSysColorBrush(COLOR_INFOBK);

    case WM_CTLCOLORSTATIC:  // Simple stunt to avoid focus rect flashing in TrackBars
        switch (GetDlgCtrlID((HWND) lParam)) {
        case IDC_TRACKBAR1:
        case IDC_TRACKBAR2:
        case IDC_TRACKBAR3:
        case IDC_TRACKBAR4:
            SetFocus(gP.hPiano);
        }
        break;

    case WM_DESTROY:
        // is sent when program ends - a good place to delete any created objects and
        // store settings in file for next run, etc. Must send PostQuitMessage to end
        // properly in SDK-style dialogs. The PostQuitMessage function sends a WM_QUIT
        // message to the program//s (thread//s) message queue, and { WM_QUIT causes
        // the GetMessage function to return zero in WINMAIN//s message loop.
        //----------------------------------------------------------------------------
        PostQuitMessage(0);
        return 0;

    case WM_HSCROLL: // Horizontal trackbar scrollpos change
        if (lParam == (LPARAM) GetDlgItem(hWnd, IDC_TRACKBAR1)) {  // Volume
            if (LOWORD(wParam) == TB_THUMBTRACK) {      // get trackbar pos
                c = (long) SendMessage((HWND) lParam, TBM_GETPOS, c, 0);
                SendMessage(gP.hPiano, PM_SETVOLUME, min(127, c * 13), 1);
            }
      
        } else if (lParam == (LPARAM) GetDlgItem(hWnd, IDC_TRACKBAR2)) { // Balance
            if (LOWORD(wParam) == TB_THUMBTRACK) {         // get trackbar pos
                c = (long) SendMessage((HWND) lParam, TBM_GETPOS, c, 0);
                SendMessage(gP.hPiano, PM_SETBALANCE, min(127, (WPARAM) (c * 6.4f + 64)), 1);
            }
      
        } else if (lParam == (LPARAM) GetDlgItem(hWnd, IDC_TRACKBAR3)) { // Modulation (Vibrato)
            if (LOWORD(wParam) == TB_THUMBTRACK) {         // get trackbar pos
                c = (long) SendMessage((HWND) lParam, TBM_GETPOS, c, 0);
                SendMessage(gP.hPiano, PM_SETVIBRATO, min(127, c * 13), 1);
            }
        }
        break;

    case WM_VSCROLL: //Pitch bend - Vertical trackbar scrollpos change
        if (lParam == (LPARAM) GetDlgItem(hWnd, IDC_TRACKBAR4)) { // Pitch Bend
            switch LOWORD(wParam) {
            case TB_THUMBTRACK: // get trackbar pos and set pitch
                c = (long) SendMessage((HWND) lParam, TBM_GETPOS, c, 0);
                SendMessage(gP.hPiano, PM_SETPITCH, 64 - c, 1);
                break;
            case TB_ENDTRACK:   // reset trackbar (middle) pos and pitch
                c = (long) SendMessage((HWND) lParam, TBM_SETPOS, TRUE, 0);
                SendMessage(gP.hPiano, PM_SETPITCH, 64, 1);
                break;
            }
        }
        break;

    case WM_HELP:
        if (F1Help == 0) {  // need a static prevention flag there, else
            F1Help = 1;     // pressing F1 again produces a new messagebox
            wcscpy_s(zText,  L"MIDI Piano accepts both computer keyboard and/or mouse play.\n\n");
            wcsncat_s(zText, L" Options:\n", _TRUNCATE);
            wcsncat_s(zText, L"   ComboBox 1 shows available midi devices (usually 1).\n", _TRUNCATE);
            wcsncat_s(zText, L"   ComboBox 2 shows a collection of 16 midi channels,\n", _TRUNCATE);
            wcsncat_s(zText, L"   but we usually only use channel 1 for normal play.\n", _TRUNCATE);
            wcsncat_s(zText, L"   Note: Channel 10 is reserved for percussion sets,\n", _TRUNCATE);
            wcsncat_s(zText, L"   turning each key from D2# to D7# into an instrument.\n\n", _TRUNCATE);
            wcsncat_s(zText, L"   Always on top - program stays topmost on screen.\n", _TRUNCATE);
            wcsncat_s(zText, L"   Show Key text - show/hide computer keyboard key text.\n", _TRUNCATE);
            wcsncat_s(zText, L"   Show Note text - show/hide piano key notes.\n\n", _TRUNCATE);
            wcsncat_s(zText, L" Effects:\n", _TRUNCATE);
            wcsncat_s(zText, L"   Arrow left - decrease octave\n", _TRUNCATE);
            wcsncat_s(zText, L"   Arrow right - increase octave\n", _TRUNCATE);
            wcsncat_s(zText, L"   Arrow up - pitch bend up\n", _TRUNCATE);
            wcsncat_s(zText, L"   Arrow down - pitch bend down\n\n", _TRUNCATE);
            wcsncat_s(zText, L"   Sustain on/off, for sustain (reverb) on played keys.\n", _TRUNCATE);
            wcsncat_s(zText, L"   Sustain can cause eternal notes to for example string and\n", _TRUNCATE);
            wcsncat_s(zText, L"   brass instruments and should therefore be used with care.\n", _TRUNCATE);
            wcsncat_s(zText, L"   if eternal notes occur, just uncheck this checkbox.\n", _TRUNCATE);
            wcsncat_s(zText, L"   On octave change, pressed notes are repeated for a\n", _TRUNCATE);
            wcsncat_s(zText, L"   nice replay effect.\n", _TRUNCATE);
            wcsncat_s(zText, L"   Mouse controls Volume, Balance, Vibrato and Pitch bend.", _TRUNCATE);
            MessageBox (hWnd, zText, L"Information", MB_OK | MB_ICONINFORMATION);
            F1Help = 0;
            break;
        }
    }

    return DefWindowProc(hWnd, uMsg, wParam, lParam);
}

int WINAPI wWinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPTSTR lpCmdLine, int nCmdShow) {
    UNREFERENCED_PARAMETER(hPrevInstance);
    UNREFERENCED_PARAMETER(lpCmdLine);
    MSG msg = { 0 };

    WCHAR zClass[] = L"MIDIpiano";
    HANDLE hMutex = CreateMutex(NULL, 0, zClass);
    if (hMutex) {
        if (GetLastError() == ERROR_ALREADY_EXISTS) {
            HWND hFound;
            for (;;) {
                hFound = FindWindow(zClass, NULL); if (hFound) { break; }
                while (PeekMessage(&msg, NULL, NULL, NULL, PM_REMOVE)) {}
            }
            if (IsIconic(hFound)) { ShowWindow(hFound, SW_RESTORE); }
            SetForegroundWindow(hFound);
            return 0;
        }
    }

    WCHAR szCaption[MAX_PATH];
    HWND hCtl = 0;
    long nRet = 0;
    WNDCLASSEX wcx = { 0 };

    //NONCLIENTMETRICS ncm = { 0 };
    //ncm.cbSize = sizeof(NONCLIENTMETRICS);
    //if (SystemParametersInfo(SPI_GETNONCLIENTMETRICS, 0, &ncm, 0)) {
    //    ncm.lfMessageFont.lfHeight = dpi(13);
    //    ncm.lfMessageFont.lfWeight = FW_EXTRABOLD; // 800;
    //    gP.usefont = CreateFontIndirect(&ncm.lfMessageFont); //Default W10 "Segoe UI"
    //} else {
        gP.usefont = (HFONT) GetStockObject(DEFAULT_GUI_FONT); // Tahoma
    //}
    gP.hinstance = hInstance;

    wcx.cbSize = sizeof(wcx);
    long IsInitialized = GetClassInfoEx(gP.hinstance, zClass, &wcx);
    if (!IsInitialized) {
        wcx.style = CS_HREDRAW | CS_VREDRAW;
        wcx.lpfnWndProc = WndProc;
        wcx.cbClsExtra = 0;
        wcx.cbWndExtra = 0;
        wcx.hInstance = gP.hinstance;
        //wcx.hIcon = LoadIcon(NULL, IDI_APPLICATION);
        wcx.hIcon = LoadIcon(wcx.hInstance, MAKEINTRESOURCE(IDI_ICON1));
        wcx.hCursor = LoadCursor(NULL, IDC_ARROW);
        wcx.hbrBackground = (HBRUSH)COLOR_BTNSHADOW;
        wcx.lpszMenuName = NULL;
        wcx.lpszClassName = zClass;
        wcx.hIconSm = wcx.hIcon;
        if (RegisterClassEx(&wcx)) { IsInitialized = -1; }
    }

    if (IsInitialized) {
        DWORD dwStyle = WS_CLIPSIBLINGS | WS_CAPTION | WS_SYSMENU | WS_MINIMIZEBOX | DS_3DLOOK | DS_NOFAILCREATE | DS_SETFONT | DS_CENTER;
        DWORD dwExStyle = WS_EX_CONTROLPARENT | WS_EX_WINDOWEDGE;
        RECT rc = { 0 };
        SetRect(&rc, 0, 0, ClientW, ClientH);
        AdjustWindowRectEx(&rc, dwStyle, FALSE, dwExStyle);

        gP.minTrackSizeW = dpi(rc.right - rc.left);
        gP.minTrackSizeH = dpi(rc.bottom - rc.top);
        long x = max((GetSystemMetrics(SM_CXSCREEN) - gP.minTrackSizeW) / 2, 0);
        long y = max((GetSystemMetrics(SM_CYSCREEN) - gP.minTrackSizeH) / 2, 0);

        // Create The Window
        wcscpy_s(szCaption, L"MIDI Piano 64-bit");
        gP.hMain = CreateWindowEx(dwExStyle, zClass, szCaption, dwStyle, x, y, gP.minTrackSizeW, gP.minTrackSizeH, 0, 0, gP.hinstance, NULL);

        InitCommonControls();

        gP.hPiano = CreatePiano(gP.hMain, IDC_PIANO, L"", 11, 193, 612, 179, WS_CHILD | WS_TABSTOP |WS_VISIBLE, 0);
        SetCtrlFont(gP.hPiano, gP.usefont);

        dwStyle = WS_CHILD | WS_VISIBLE | BS_GROUPBOX;
        hCtl = CreateWindowEx(0, L"BUTTON", L"Options", dwStyle, 8 + 3, 3, 195, 176, gP.hMain, (HMENU) IDC_FRAME1, gP.hinstance, NULL);
        SetCtrlFont(hCtl, gP.usefont);

        dwStyle =  WS_CHILD | WS_VISIBLE | WS_TABSTOP | CBS_DROPDOWNLIST | CBS_HASSTRINGS;
        hCtl = CreateWindowEx(0, L"COMBOBOX", NULL, dwStyle, 17 + 3, 21, 177, 160, gP.hMain, (HMENU) IDC_COMBOBOX1, gP.hinstance, NULL);
        SetCtrlFont(hCtl, gP.usefont);
        SendMessage(gP.hPiano, PM_LISTDEVICES, IDC_COMBOBOX1, 0);

        dwStyle = WS_CHILD | WS_VISIBLE | WS_TABSTOP | CBS_DROPDOWNLIST | CBS_HASSTRINGS;
        hCtl = CreateWindowEx(0, L"COMBOBOX", NULL, dwStyle, 17 + 3, 52, 177, 240, gP.hMain, (HMENU) IDC_COMBOBOX2, gP.hinstance, NULL);
        SetCtrlFont(hCtl, gP.usefont);
        SendMessage(gP.hPiano, PM_LISTCHANNELS, IDC_COMBOBOX2, 0);

        dwStyle = WS_CHILD | WS_VISIBLE | WS_TABSTOP | BS_AUTOCHECKBOX;
        hCtl = CreateWindowEx(0, L"BUTTON", L"Always on top ", dwStyle, 17 + 3, 85, 110, 16, gP.hMain, (HMENU) IDC_CHECKBOX1, gP.hinstance, NULL);
        SetCtrlFont(hCtl, gP.usefont);

        dwStyle = WS_CHILD | WS_VISIBLE | WS_TABSTOP | BS_AUTOCHECKBOX;
        hCtl = CreateWindowEx(0, L"BUTTON", L"Show Key text ", dwStyle, 17 + 3, 104, 110, 16, gP.hMain, (HMENU) IDC_CHECKBOX2, gP.hinstance, NULL);
        SetCtrlFont(hCtl, gP.usefont);
        SendMessage(hCtl, BM_SETCHECK, BST_CHECKED, 0);

        dwStyle = WS_CHILD | WS_VISIBLE | WS_TABSTOP | BS_AUTOCHECKBOX;
        hCtl = CreateWindowEx(0, L"BUTTON", L"Show Note text", dwStyle, 17 + 3, 124, 110, 16, gP.hMain, (HMENU) IDC_CHECKBOX3, gP.hinstance, NULL);
        SetCtrlFont(hCtl, gP.usefont);
        SendMessage(hCtl, BM_SETCHECK, BST_CHECKED, 0);

        dwStyle = WS_CHILD | WS_VISIBLE | SS_CENTER;
        hCtl = CreateWindowEx(0, L"STATIC", L"Press F1 for help, Esc to Exit", dwStyle, 17 + 3, 159, 177, 17, gP.hMain, (HMENU) IDC_LABEL4, gP.hinstance, NULL);
        SetCtrlFont(hCtl, gP.usefont);

        dwStyle = WS_CHILD | WS_VISIBLE | BS_GROUPBOX;
        hCtl = CreateWindowEx(0, L"BUTTON", L"Patches/Instrument", dwStyle, 215, 3, 160, 176, gP.hMain, (HMENU) IDC_FRAME2, gP.hinstance, NULL);
        SetCtrlFont(hCtl, gP.usefont);

        dwStyle = WS_CHILD | WS_VISIBLE | WS_TABSTOP | CBS_DROPDOWNLIST | CBS_HASSTRINGS;
        hCtl = CreateWindowEx(0, L"COMBOBOX", NULL, dwStyle, 224, 20, 142, 260, gP.hMain, (HMENU) IDC_COMBOBOX3, gP.hinstance, NULL);
        SetCtrlFont(hCtl, gP.usefont);
        SendMessage(gP.hPiano, PM_LISTPATCHES, IDC_COMBOBOX3, 0);
        SendMessage(hCtl, CB_SETCURSEL, 0, 0);

        dwStyle = WS_CHILD | WS_VISIBLE | WS_TABSTOP | LBS_NOTIFY | LBS_NOINTEGRALHEIGHT;
        hCtl = CreateWindowEx(WS_EX_CLIENTEDGE, L"LISTBOX", NULL, dwStyle, 224, 47 - 5, 142, 124 + 6, gP.hMain, (HMENU) IDC_LISTBOX1, gP.hinstance, NULL);
        SetCtrlFont(hCtl, gP.usefont);
        SendMessage(gP.hPiano, PM_LISTINSTRUMENTS, IDC_LISTBOX1, 0);
        SendMessage(hCtl, LB_SETCURSEL, 0, 0);

        dwStyle = WS_CHILD | WS_VISIBLE | BS_GROUPBOX;
        hCtl = CreateWindowEx(0, L"BUTTON", L"Effects", dwStyle, 387 - 3, 3, 156, 176, gP.hMain, (HMENU) IDC_FRAME3, gP.hinstance, NULL);
        SetCtrlFont(hCtl, gP.usefont);

        dwStyle = WS_CHILD | WS_VISIBLE | SS_SUNKEN;
        hCtl = CreateWindowEx(WS_EX_STATICEDGE, L"STATIC", NULL, dwStyle, 398 - 3, 20, 135, 30, gP.hMain, (HMENU) -1, gP.hinstance, NULL);
        SetCtrlFont(hCtl, gP.usefont);

        dwStyle = WS_CHILD | WS_VISIBLE | SS_CENTER;
        hCtl = CreateWindowEx(0, L"STATIC", L"Octave 4 to 5 \n(< or > to change)", dwStyle, 416 - 3, 21, 97, 28, gP.hMain, (HMENU) IDC_LABEL1, gP.hinstance, NULL);
        SetCtrlFont(hCtl, gP.usefont);

        dwStyle = WS_CHILD | WS_VISIBLE;// | SS_CENTERIMAGE | SS_NOTIFY | SS_CENTER;
        hCtl = CreateWindowEx(0, L"BUTTON", L"<", dwStyle, 399 - 3, 21, 18, 28, gP.hMain, (HMENU) IDC_LABEL2, gP.hinstance, NULL);
        SetCtrlFont(hCtl, gP.usefont);

        dwStyle = WS_CHILD | WS_VISIBLE;//| SS_CENTERIMAGE | SS_NOTIFY | SS_CENTER;
        hCtl = CreateWindowEx(0, L"BUTTON", L">", dwStyle, 513 - 3, 21, 18, 28, gP.hMain, (HMENU) IDC_LABEL3, gP.hinstance, NULL);
        SetCtrlFont(hCtl, gP.usefont);

        dwStyle = WS_CHILD | WS_VISIBLE;
        hCtl = CreateWindowEx(0, L"STATIC", L"Volume", dwStyle, 395 - 3, 60 + 3, 45, 16, gP.hMain, (HMENU) -1, gP.hinstance, NULL);
        SetCtrlFont(hCtl, gP.usefont);

        dwStyle = WS_CHILD | WS_VISIBLE | TBS_AUTOTICKS | TBS_BOTTOM | TBS_TOOLTIPS;
        hCtl = CreateWindowEx(0, L"msctls_trackbar32", NULL, dwStyle, 443 - 4, 56, 96, 32, gP.hMain, (HMENU) IDC_TRACKBAR1, gP.hinstance, NULL);
        SetCtrlFont(hCtl, gP.usefont);
        SendMessage(hCtl, TBM_SETRANGE , TRUE, MAKLNG(0, 10));
        SendMessage(hCtl, TBM_SETTICFREQ, 1, 1); // interval frequency
        SendMessage(hCtl, TBM_SETPOS, TRUE, 10); // position

        dwStyle = WS_CHILD | WS_VISIBLE;
        hCtl = CreateWindowEx(0, L"STATIC", L"Balance", dwStyle, 395 - 3, 93 + 3, 45, 16, gP.hMain, (HMENU) -1, gP.hinstance, NULL);
        SetCtrlFont(hCtl, gP.usefont);

        dwStyle = WS_CHILD | WS_VISIBLE | TBS_AUTOTICKS | TBS_BOTTOM | TBS_TOOLTIPS;
        hCtl = CreateWindowEx(0, L"msctls_trackbar32", NULL, dwStyle, 443 - 4, 88, 96, 33, gP.hMain, (HMENU) IDC_TRACKBAR2, gP.hinstance, NULL);
        SetCtrlFont(hCtl, gP.usefont);
        SendMessage(hCtl, TBM_SETRANGE, TRUE, MAKLNG(-10, 10));
        SendMessage(hCtl, TBM_SETTICFREQ, 5, 1); 
        SendMessage(hCtl, TBM_SETPOS, TRUE, 0); // balance center

        dwStyle = WS_CHILD | WS_VISIBLE;
        hCtl = CreateWindowEx(0, L"STATIC", L"Vibrato", dwStyle, 395 - 3, 125 + 3, 45, 16, gP.hMain, (HMENU) -1, gP.hinstance, NULL);
        SetCtrlFont(hCtl, gP.usefont);

        dwStyle = WS_CHILD | WS_VISIBLE | TBS_AUTOTICKS | TBS_BOTTOM | TBS_TOOLTIPS;
        hCtl = CreateWindowEx(0, L"msctls_trackbar32", NULL, dwStyle, 443 - 4, 121, 96, 32, gP.hMain, (HMENU) IDC_TRACKBAR3, gP.hinstance, NULL);
        SetCtrlFont(hCtl, gP.usefont);
        SendMessage(hCtl, TBM_SETRANGE, TRUE, MAKLNG(0, 10));
        SendMessage(hCtl, TBM_SETTICFREQ, 1, 1); // interval frequency
        SendMessage(hCtl, TBM_SETPOS, TRUE, 0);  // position

        dwStyle = WS_CHILD | WS_VISIBLE | WS_TABSTOP | BS_AUTOCHECKBOX;
        hCtl = CreateWindowEx(0, L"BUTTON", L"Sustain", dwStyle, 396 - 3, 156, 65, 16, gP.hMain, (HMENU) IDC_CHECKBOX4, gP.hinstance, NULL);
        SetCtrlFont(hCtl, gP.usefont);

        dwStyle = WS_CHILD | WS_VISIBLE | BS_GROUPBOX;
        hCtl = CreateWindowEx(0, L"BUTTON", L"Pitch bend", dwStyle, 555 - 6, 3, 68 + 6, 176, gP.hMain, (HMENU) IDC_FRAME4, gP.hinstance, NULL);
        SetCtrlFont(hCtl, gP.usefont);

        dwStyle = WS_CHILD | WS_VISIBLE | TBS_AUTOTICKS | TBS_VERT | TBS_RIGHT | TBS_BOTH;
        hCtl = CreateWindowEx(0, L"msctls_trackbar32", NULL, dwStyle, 569 - 5, 20, 45, 152, gP.hMain, (HMENU) IDC_TRACKBAR4, gP.hinstance, NULL);
        SetCtrlFont(hCtl, gP.usefont);
        SendMessage(hCtl, TBM_SETRANGE , TRUE, MAKLNG(-64, 64));
        SendMessage(hCtl, TBM_SETTICFREQ, 10, 1);

        // Show the main window
        ShowWindow(gP.hMain, nCmdShow);
        UpdateWindow(gP.hMain);

        while (GetMessage(&msg, NULL, 0, 0)) {
            TranslateMessage(&msg);
            DispatchMessage(&msg);
        }
    }

    if (hMutex) { CloseHandle(hMutex); }
    return nRet;
}
