#pragma once
// Piano.h - Public Domain by Borje Hagsten, June, 2016.
// C++ version (VS Community 2015) by Patrice Terrier, June, 2016.
//
// To begin with, computer keyboards are lousy midi keyboards. Some
// key combinations when playing chords simply don//t work, probably
// because of bit mask conflicts. But it//s fun to learn new stuff
// and investigate, which is why I did this one. Pitch bending and
// vibrato were the hardest part to do, because I could not find any
// examples or good info how to do them by code, so had to experiment
// and finally found a way. Should be easy enough to understand how
// to use all features and learn limitations. Both mouse and keyboard
// can be used to play at same time.
//
// As always, code is free to use and abuse any way you want. Could
// serve as base for a note learning game or a sequenser, build own
// custom signals or just use it as it is, for fun.

#include <Mmsystem.h>

#define PM_SETDEVICE       WM_APP + 100  // wParam sets midi device
#define PM_SETCHANNEL      WM_APP + 101  // wParam sets channel
#define PM_PLAYNOTE        WM_APP + 105  // wParam sets note to play (0 to 127), lParam sets sound volume (0 to 127)
#define PM_STOPNOTE        WM_APP + 106  // wParam determines which note to stop playing (0 to 127)

#define PM_LISTDEVICES     WM_APP + 110  // List midi devices. wParam is list id
#define PM_LISTCHANNELS    WM_APP + 111  // List 16 midi channels. wParam list id
#define PM_LISTPATCHES     WM_APP + 112  // List patches. wParam is list id
#define PM_LISTINSTRUMENTS WM_APP + 113  // List instrument. wParam is list id

// For the following #define PM_SET.. messages, lParam (1/0) sets control focus
#define PM_SETINSTRUMENT   WM_APP + 120  // wParam sets instrument (0-127)
#define PM_SETOCTAVE       WM_APP + 121  // wParam sets octave (0-10)
#define PM_SETKEYCOUNT     WM_APP + 122  // wParam sets total number of keys (1 to 128)
#define PM_SETPITCH        WM_APP + 123  // wParam sets pitch bend (-64 to 64, where 0 is normal pitch)
#define PM_SETVOLUME       WM_APP + 124  // wParam sets sound volume (0 to 127)
#define PM_SETBALANCE      WM_APP + 125  // wParam sets sound balance (-64 to 64, where 0 is centered)
#define PM_SETVIBRATO      WM_APP + 126  // wParam sets vibrato/modulation (0 to 127)
#define PM_SETSUSTAIN      WM_APP + 127  // wParam sets sustain off/on (0/127)
#define PM_SETKEYTEXT      WM_APP + 130  // wParam 0/1, if to print computer key letters
#define PM_SETNOTETEXT     WM_APP + 131  // wParam 0/1, if to print note desciptions

#define PM_GETOCTAVE       WM_APP + 221  // get base octave
#define PM_GETKEYCOUNT     WM_APP + 222  // get number of keys (1 to 128)

#define PN_PLAYNOTE        WM_APP + 300  // note played notification
#define PN_STOPNOTE        WM_APP + 301  // note stopped notification
#define PN_OCTAVE          WM_APP + 310  // octave change notification
#define PN_PITCHBEND       WM_APP + 320  // pitch bend notification

#define VK_0               0x30
#define VK_1               0x31
#define VK_2               0x32
#define VK_3               0x33
#define VK_4               0x34
#define VK_5               0x35
#define VK_6               0x36
#define VK_7               0x37
#define VK_8               0x38
#define VK_9               0x39
#define VK_A               0x41
#define VK_B               0x42
#define VK_C               0x43
#define VK_D               0x44
#define VK_E               0x45
#define VK_F               0x46
#define VK_G               0x47
#define VK_H               0x48
#define VK_I               0x49
#define VK_J               0x4A
#define VK_K               0x4B
#define VK_L               0x4C
#define VK_M               0x4D
#define VK_N               0x4E
#define VK_O               0x4F
#define VK_P               0x50
#define VK_Q               0x51
#define VK_R               0x52
#define VK_S               0x53
#define VK_T               0x54
#define VK_U               0x55
#define VK_V               0x56
#define VK_W               0x57
#define VK_X               0x58
#define VK_Y               0x59
#define VK_Z               0x5A

#define BLACK              0x000000
#define WHITE              0xFFFFFF
#define LTGRAY             0xC0C0C0
#define GRAY               0x808080
#define YELLOW             0x00FFFF
#define BLUE               0xFF0000
#define RED                0x0000FF

const int UB = 128;
#define GWL_HINSTANCE      (-6)

//--------------------------------------------------------------------
struct kbData {     // Program data
  HWND hParent;     // Parent dialog handle
  long CtlId;       // Piano control id
  HFONT hFont;      // font handle
  HMIDIOUT hMidi;   // midi device
  HDC memDC;        // memory device context
  HBITMAP hBitmap;  //
  HBITMAP hBitOld;  //
  long KeyCount;    // Number of keys to draw
  long instrument;  // selected instrument (0 - 127)
  long chPlay;      // channel &H90 to &H9F, where ch &H99 = percussion
  long chStop;      // channel &H80 to &H8F, where ch &H89 = percussion
  long chControl;   // channel &HB0 to &HBF, Control mode change
  long chInstr;     // channel &HC0 to &HCF, Program (instrument) change
  long chBend;      // channel &HE0 to &HEF, Pitch bend
  long KeyText;     // Key description on keys on/off
  long NoteText;    // Note description on keys on/off
  long note;        // note to play (0 - 127)
  long octave;      // base octave (1 octave = 12 notes)
  long sustain;     // sustain on/off
  long volume;      // sound volume
  long balance;     // sound balance Left/Right
  long vibrato;     // Modulation/Vibrato
};

struct pKey {       // Key array data
  long nt;          // note (0 - 127)
  long dn;          // key up/down flag, 0/1
  RECT rc;          // key coordinates
};

WCHAR KEY1[] = { L"ZSXDCVGBHNJMQ2W3ER5T6Y7UI9O0P" };  // 29 computer key note layout
kbData g_kbd;    // static keyboard dat
HRGN hRgn[UB];   // static key region ar
pKey pk[UB];     // static key data arra

HFONT MakeFontEx(IN HDC hDC, IN WCHAR* zFontName, IN long PointSize, IN long fAngle, IN long fBold, IN long fItalic, IN long fUnderline) {
    LPARAM CharSet = 0; long CyPixels = 0;
    if (hDC == 0) {
        hDC = GetDC(HWND_DESKTOP);
        CyPixels  = GetDeviceCaps(hDC, LOGPIXELSY);
        ReleaseDC(HWND_DESKTOP, hDC);
    } else {
        CyPixels  = GetDeviceCaps(hDC, LOGPIXELSY);
    }
    PointSize = 0 - (PointSize * CyPixels) / 72;
    return CreateFont(PointSize, 0, fAngle, fAngle, fBold, fItalic, fUnderline, FALSE,
                      DEFAULT_CHARSET, OUT_TT_PRECIS, CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY, FF_DONTCARE, zFontName);
}

// Set desired instrument, event/channel 0xB0 to 0xBF.
MMRESULT Midi_SetInstrument(IN HMIDIOUT hMidi, IN long nEvent, IN long nInstrument) { // instrument
    DWORD dwMsg = nEvent + nInstrument * 0x100; // setup dwMsg
    return midiOutShortMsg(hMidi, dwMsg); // set instrument
}

// Set sound volume for selected channel, event/channel 0xC0 to 0xCF.
MMRESULT Midi_SetVolume(IN HMIDIOUT hMidi, IN long nEvent, IN long nVolume) {
    DWORD dwMsg = nEvent + 0x7 * 0x100 + nVolume * 0x10000; // 0x7 = 7
    return midiOutShortMsg(hMidi, dwMsg);
}

// Set Modulation/Vibrato for selected channel. iModulation = 1 to 127
MMRESULT Midi_SetModulation(IN HMIDIOUT hMidi, IN long nEvent, IN long nModulation) {
    DWORD dwMsg = nEvent + 0x1 * 0x100 + nModulation * 0x10000;  // 0x1 = 1
    return midiOutShortMsg(hMidi, dwMsg);
}

// Pitch Bend +/- 2 notes, 0 = Max down, 64=Normal note, 127 = Max up
MMRESULT Midi_PitchBend (IN HMIDIOUT hMidi, IN long nEvent, IN long nPitch) {
    DWORD dwMsg = nEvent + nPitch * 0x10000;  // setup dwMsg
    return midiOutShortMsg(hMidi, dwMsg);
}

// Set balance for selected channel.  0 = Left, 64 = Center, 127 = Right
MMRESULT Midi_SetBalance(IN HMIDIOUT hMidi, IN long nEvent, IN long nBalance) {
    DWORD dwMsg = nEvent + 0xA * 0x100 + nBalance * 0x10000; // 0xA = 10
    return midiOutShortMsg(hMidi, dwMsg);
}

// Set Sustain for selected channel
MMRESULT Midi_SetSustain(IN HMIDIOUT hMidi, IN long nEvent, IN long nSustain) {
    DWORD dwMsg = nEvent + 0x40 * 0x100 + nSustain * 0x10000;  // 0x40 = 64
    return midiOutShortMsg(hMidi, dwMsg);
}

// Play a note using midiOutShortMsg, channel 0x90 to 0x9F (ch 0x99 = percussion)
MMRESULT Midi_PlayNote (IN HMIDIOUT hMidi, IN long nEvent, IN long nNote, IN long nVolume) {
    DWORD dwMsg = nEvent + nNote * 0x100 + nVolume * 0x10000; // setup for playing
    return midiOutShortMsg(hMidi, dwMsg);                     // play note
}

// Turn off all notes in selected channel, 0xB0 to 0xBF.
MMRESULT Midi_StopAllNotes(IN HMIDIOUT hMidi, IN long nEvent) { //event/channel
    DWORD dwMsg = nEvent + 0x7B * 0x100;  // 0x7B = 123
    return midiOutShortMsg(hMidi, dwMsg);
}

// Stop playing a note using midiOutShortMsg, channel 0x80 to 0x8F.
MMRESULT Midi_StopNote (IN HMIDIOUT hMidi, IN long nEvent, IN long nNote) {
    DWORD dwMsg = nEvent + (nNote * 0x100); // setup dwMsg
    return midiOutShortMsg(hMidi, dwMsg);   // stop playing
}

// Get note description for 7 + 5  =12 keys in any octave-
WCHAR* GetNoteText(IN long iNote) {
    switch (iNote % 12) {
    case  0 : return L"C";
    case  1 : return L"C#";
    case  2 : return L"D";
    case  3 : return L"D#";
    case  4 : return L"E";
    case  5 : return L"F";
    case  6 : return L"F#";
    case  7 : return L"G";
    case  8 : return L"G#";
    case  9 : return L"A";
    case 10 : return L"A#";
    case 11 : return L"B";
    }
}

// Print note + key text on key, using desired colors.
void PrintKeyText(pKey pk, IN long iForeColor1, IN long iBackColor1, IN long iForeColor2, IN long iBackColor2) {

    WCHAR zText[3] = { 0 };
    RECT rc = { 0 }; SIZEL sz = { 0 };

    HFONT tmpFont = (HFONT) SelectObject(g_kbd.memDC, g_kbd.hFont);
    GetTextExtentPoint32(g_kbd.memDC, L"W", 1, &sz);  // need font height
    rc = pk.rc;
    memset(&zText[0], 0, sizeof(zText));
    CopyMemory(&zText[0], &KEY1[pk.nt], 2);
    SetBkColor(g_kbd.memDC, iBackColor1);
    SetTextColor(g_kbd.memDC, iForeColor1);
    rc.bottom = (long) (pk.rc.bottom - 1.7f * sz.cy);
    DrawText(g_kbd.memDC, zText, (long) wcslen(zText), &rc, DT_BOTTOM | DT_CENTER | DT_SINGLELINE);

    // draw note text
    switch (pk.nt % 12) {
    case  0 : wcscpy_s(zText, L"C"); break;
    case  1 : wcscpy_s(zText, L"C#"); break;
    case  2 : wcscpy_s(zText, L"D"); break;
    case  3 : wcscpy_s(zText, L"D#"); break;
    case  4 : wcscpy_s(zText, L"E"); break;
    case  5 : wcscpy_s(zText, L"F"); break;
    case  6 : wcscpy_s(zText, L"F#"); break;
    case  7 : wcscpy_s(zText, L"G"); break;
    case  8 : wcscpy_s(zText, L"G#"); break;
    case  9 : wcscpy_s(zText, L"A"); break;
    case 10 : wcscpy_s(zText, L"A#"); break;
    case 11 : wcscpy_s(zText, L"B"); break;
    }

    SetBkColor(g_kbd.memDC, iBackColor2);
    SetTextColor(g_kbd.memDC, iForeColor2);
    rc.bottom = (long) (pk.rc.bottom - 0.5f * sz.cy);
    DrawText(g_kbd.memDC, zText, (long) wcslen(zText), &rc, DT_BOTTOM | DT_CENTER | DT_SINGLELINE);

    if (tmpFont) { SelectObject(g_kbd.memDC, tmpFont); }  // restore font
}

// Enumerate midi devices - cannot test if it works 100%, but think it is ok..
void Midi_ListDevices (IN HWND hDlg, IN long CtlId) {
    long c = 0; HWND hCtrl = 0;
    DWORD dwMsg1 = 0, dwMsg2 = 0;
    MIDIOUTCAPS MidiCaps = { 0 };
    WCHAR zTxt[32];

    hCtrl = GetDlgItem(hDlg, CtlId);
    GetClassName(hCtrl, zTxt, sizeof(zTxt)); // see what kind of control to fill
    CharUpper(zTxt);
    if (_wcsicmp(zTxt, L"COMBOBOX") == 0) {        // if to add results to a ComboBox
        SendMessage(hCtrl, CB_RESETCONTENT, 0, 0);
        dwMsg1 = CB_ADDSTRING;
        dwMsg2 = CB_SETCURSEL;
    } else {                                             // elseif to add results to a ListBox
        SendMessage(hCtrl, LB_RESETCONTENT, 0, 0);
        dwMsg1 = LB_ADDSTRING;
        dwMsg2 = LB_SETCURSEL;
    }
    //------------------------------------------------------------------
    for (UINT dev = 0; dev < midiOutGetNumDevs(); dev++) {  // there is usually only one...
        midiOutGetDevCaps(dev, &MidiCaps, sizeof(MidiCaps));
        SendMessage(hCtrl, dwMsg1, 0, (LPARAM) MidiCaps.szPname);
    }
    SendMessage(hCtrl, dwMsg2, 0, 0);
}

// Fill a ListBox or ComboBox with midi instruments
void Midi_ListInstruments(IN HWND hDlg, IN long CtlId, IN long iPatches) {
    long c = 0, iStart = 0, iStop = 0;
    HWND hCtrl = 0;
    DWORD dwMsg1 = 0, dwMsg2 = 0;
    WCHAR zTxt[64] = { 0 };

    static HWND hList;
    if (hList == 0) {
        HINSTANCE hInstance = (HINSTANCE) GetWindowLongPtr(hDlg, GWL_HINSTANCE);
        hList = CreateWindowEx(0, L"LISTBOX", NULL, WS_CHILD, 0, -1000, 0, 0, hDlg, 0, hInstance, NULL);
        if (hList) {
            ListAdd(hList, L"1 Acoustic Grand Piano");
            ListAdd(hList, L"2 Bright Acoustic Piano");
            ListAdd(hList, L"3 Electric Grand Piano");
            ListAdd(hList, L"4 Honky-tonk Piano");
            ListAdd(hList, L"5 Rhodes Piano");
            ListAdd(hList, L"6 Chorused Piano");
            ListAdd(hList, L"7 Harpsichord");
            ListAdd(hList, L"8 Clavinet");
            ListAdd(hList, L"9 Celesta");
            ListAdd(hList, L"10 GlockenspieL");
            ListAdd(hList, L"11 Music Box");
            ListAdd(hList, L"12 Vibraphone");
            ListAdd(hList, L"13 Marimba");
            ListAdd(hList, L"14 Xylophone");
            ListAdd(hList, L"15 Tubular Bells");
            ListAdd(hList, L"16 Dulcimer");
            ListAdd(hList, L"17 Hammond Organ");
            ListAdd(hList, L"18 Percussive Organ");
            ListAdd(hList, L"19 Rock Organ");
            ListAdd(hList, L"20 Church Organ");
            ListAdd(hList, L"21 Reed Organ");
            ListAdd(hList, L"22 Accordion");
            ListAdd(hList, L"23 Harmonica");
            ListAdd(hList, L"24 Tango Accordion");
            ListAdd(hList, L"25 Acoustic Guitar (nylon)");
            ListAdd(hList, L"26 Acoustic Guitar (steel)");
            ListAdd(hList, L"27 Electric Guitar (jazz)");
            ListAdd(hList, L"28 Electric Guitar (clean)");
            ListAdd(hList, L"29 Electric Guitar (muted)");
            ListAdd(hList, L"30 Overdriven Guitar");
            ListAdd(hList, L"31 Distortion Guitar");
            ListAdd(hList, L"32 Guitar Harmonics");
            ListAdd(hList, L"33 Acoustic Bass");
            ListAdd(hList, L"34 Electric Bass (finger)");
            ListAdd(hList, L"35 Electric Bass (pick)");
            ListAdd(hList, L"36 Fretless Bass");
            ListAdd(hList, L"37 Slap Bass 1");
            ListAdd(hList, L"38 Slap Bass 2");
            ListAdd(hList, L"39 Synth Bass 1");
            ListAdd(hList, L"40 Synth Bass 2");
            ListAdd(hList, L"41 Violin");
            ListAdd(hList, L"42 Viola");
            ListAdd(hList, L"43 Cello");
            ListAdd(hList, L"44 Contrabass");
            ListAdd(hList, L"45 Tremolo Strings");
            ListAdd(hList, L"46 Pizzicato Strings");
            ListAdd(hList, L"47 Orchestral Harp");
            ListAdd(hList, L"48 Timpani");
            ListAdd(hList, L"49 String Ensemble 1");
            ListAdd(hList, L"50 String Ensemble 2");
            ListAdd(hList, L"51 SynthStrings 1");
            ListAdd(hList, L"52 SynthStrings 2");
            ListAdd(hList, L"53 Choir Aahs");
            ListAdd(hList, L"54 Voice Oohs");
            ListAdd(hList, L"55 Synth Voice");
            ListAdd(hList, L"56 Orchestra Hit");
            ListAdd(hList, L"57 Trumpet");
            ListAdd(hList, L"58 Trombone");
            ListAdd(hList, L"59 Tuba");
            ListAdd(hList, L"60 Muted Trumpet");
            ListAdd(hList, L"61 French Horn");
            ListAdd(hList, L"62 Brass Section");
            ListAdd(hList, L"63 Synth Brass 1");
            ListAdd(hList, L"64 Synth Brass 2");
            ListAdd(hList, L"65 Soprano Sax");
            ListAdd(hList, L"66 Alto Sax");
            ListAdd(hList, L"67 Tenor Sax");
            ListAdd(hList, L"68 Baritone Sax");
            ListAdd(hList, L"69 Oboe");
            ListAdd(hList, L"70 English Horn");
            ListAdd(hList, L"71 Bassoon");
            ListAdd(hList, L"72 Clarinet");
            ListAdd(hList, L"73 Piccolo");
            ListAdd(hList, L"74 Flute");
            ListAdd(hList, L"75 Recorder");
            ListAdd(hList, L"76 Pan Flute");
            ListAdd(hList, L"77 Bottle Blow");
            ListAdd(hList, L"78 Shakuhachi");
            ListAdd(hList, L"79 Whistle");
            ListAdd(hList, L"80 Ocarina");
            ListAdd(hList, L"81 Lead 1 (square)");
            ListAdd(hList, L"82 Lead 2 (sawtooth)");
            ListAdd(hList, L"83 Lead 3 (calliope lead)");
            ListAdd(hList, L"84 Lead 4 (chiff lead)");
            ListAdd(hList, L"85 Lead 5 (charang)");
            ListAdd(hList, L"86 Lead 6 (voice)");
            ListAdd(hList, L"87 Lead 7 (fifths)");
            ListAdd(hList, L"88 Lead 8 (bass + lead)");
            ListAdd(hList, L"89 Pad 1 (new age)");
            ListAdd(hList, L"90 Pad 2 (warm)");
            ListAdd(hList, L"91 Pad 3 (polysynth)");
            ListAdd(hList, L"92 Pad 4 (choir)");
            ListAdd(hList, L"93 Pad 5 (bowed)");
            ListAdd(hList, L"94 Pad 6 (metallic)");
            ListAdd(hList, L"95 Pad 7 (halo)");
            ListAdd(hList, L"96 Pad 8 (sweep)");
            ListAdd(hList, L"97 FX 1 (rain)");
            ListAdd(hList, L"98 FX 2 (soundtrack)");
            ListAdd(hList, L"99 FX 3 (crystal)");
            ListAdd(hList, L"100 FX 4 (atmosphere)");
            ListAdd(hList, L"101 FX 5 (brightness)");
            ListAdd(hList, L"102 FX 6 (goblins)");
            ListAdd(hList, L"103 FX 7 (echoes)");
            ListAdd(hList, L"104 FX 8 (sci-fi)");
            ListAdd(hList, L"105 Sitar");
            ListAdd(hList, L"106 Banjo");
            ListAdd(hList, L"107 Shamisen");
            ListAdd(hList, L"108 Koto");
            ListAdd(hList, L"109 Kalimba");
            ListAdd(hList, L"110 Bagpipe");
            ListAdd(hList, L"111 Fiddle");
            ListAdd(hList, L"112 Shanai");
            ListAdd(hList, L"113 Tinkle BelL");
            ListAdd(hList, L"114 Agogo");
            ListAdd(hList, L"115 Steel Drums");
            ListAdd(hList, L"116 Woodblock");
            ListAdd(hList, L"117 Taiko Drum");
            ListAdd(hList, L"118 Melodic Tom");
            ListAdd(hList, L"119 Synth Drum");
            ListAdd(hList, L"120 Reverse CymbaL");
            ListAdd(hList, L"121 Guitar Fret Noise");
            ListAdd(hList, L"122 Breath Noise");
            ListAdd(hList, L"123 Seashore");
            ListAdd(hList, L"124 Bird Tweet");
            ListAdd(hList, L"125 Telephone Ring");
            ListAdd(hList, L"126 Helicopter");
            ListAdd(hList, L"127 Applause");
            ListAdd(hList, L"128 Gunshot");
        }
    }

    switch (iPatches) {  // divide into iPatch categories
    case  0: iStart =   1; iStop =   8;  // Piano
             break;
    case  1: iStart =   9; iStop =  16;  // Chromatic Percussion
             break;
    case  2: iStart =  17; iStop =  24;  // Organ
             break;
    case  3: iStart =  25; iStop =  32;  // Guitar
             break;
    case  4: iStart =  33; iStop =  40;  // Bass
             break;
    case  5: iStart =  41; iStop =  48;  // Strings
             break;
    case  6: iStart =  49; iStop =  56;  // Ensamble
             break;
    case  7: iStart =  57; iStop =  64;  // Brass
             break;
    case  8: iStart =  65; iStop =  72;  // Reed
             break;
    case  9: iStart =  73; iStop =  80;  // Pipe
             break;
    case 10: iStart =  81; iStop =  88;  // Synth Lead
             break;
    case 11: iStart =  89; iStop =  96;  // Synth Pad
             break;
    case 12: iStart =  97; iStop = 104;  // Synth Effects
             break;
    case 13: iStart = 105; iStop = 112;  // Ethnic
             break;
    case 14: iStart = 113; iStop = 119;  // Percussive
             break;
    case 15: iStart = 120; iStop = 128;  // Sound effects
             break;
    }

    hCtrl = GetDlgItem(hDlg, CtlId);
    GetClassName(hCtrl, zTxt, sizeof(zTxt)); // see what kind of control to fill
    CharUpper(zTxt);
    if (_wcsicmp(zTxt, L"LISTBOX") == 0) {         // if to add data to a ListBox
        SendMessage(hCtrl, LB_RESETCONTENT, 0, 0);
        dwMsg1 = LB_ADDSTRING;
        dwMsg2 = LB_SETITEMDATA;
    } else if (_wcsicmp(zTxt, L"COMBOBOX") == 0) {  // elseif to add data to a ComboBox
        SendMessage(hCtrl, CB_RESETCONTENT, 0, 0);
        dwMsg1 = CB_ADDSTRING;
        dwMsg2 = CB_SETITEMDATA;
    }

    if (IsWindow(hList)) {
        for (c = iStart; c <= iStop; c++) {
            SendMessage(hCtrl, dwMsg1, 0, (LPARAM) ListGetText(hList, c));
            SendMessage(hCtrl, dwMsg2, c - iStart, c - 1);
        }
    }
}

// Fill a ComboBox or ListBox with midi patches (instrument categories)
void Midi_ListPatches(IN HWND hDlg, IN long CtlId) {

    long c = 0; HWND hCtrl = 0; DWORD dwMsg;
    WCHAR zTxt[64] = { 0 };

    hCtrl = GetDlgItem(hDlg, CtlId);
    if (IsWindow(hCtrl)) {
        GetClassName(hCtrl, zTxt, sizeof(zTxt)); // see what kind of control to fill
        CharUpper(zTxt);
        if (_wcsicmp(zTxt, L"COMBOBOX") == 0){        // if to add data to a ComboBox
            SendMessage(hCtrl, CB_RESETCONTENT, 0, 0);
            dwMsg = CB_ADDSTRING;
        } else if (_wcsicmp(zTxt, L"LISTBOX") == 0) {   // elseif to add data to a ListBox
            SendMessage(hCtrl, LB_RESETCONTENT, 0, 0);
            dwMsg = LB_ADDSTRING;
        }
        SendMessage(hCtrl, dwMsg, 0, (LPARAM) L"Piano");
        SendMessage(hCtrl, dwMsg, 0, (LPARAM) L"Chromatic Percussion");
        SendMessage(hCtrl, dwMsg, 0, (LPARAM) L"Organ");
        SendMessage(hCtrl, dwMsg, 0, (LPARAM) L"Guitar");
        SendMessage(hCtrl, dwMsg, 0, (LPARAM) L"Bass");
        SendMessage(hCtrl, dwMsg, 0, (LPARAM) L"Strings");
        SendMessage(hCtrl, dwMsg, 0, (LPARAM) L"Ensamble");
        SendMessage(hCtrl, dwMsg, 0, (LPARAM) L"Brass");
        SendMessage(hCtrl, dwMsg, 0, (LPARAM) L"Reed");
        SendMessage(hCtrl, dwMsg, 0, (LPARAM) L"Pipe");
        SendMessage(hCtrl, dwMsg, 0, (LPARAM) L"Synth Lead");
        SendMessage(hCtrl, dwMsg, 0, (LPARAM) L"Synth Pad");
        SendMessage(hCtrl, dwMsg, 0, (LPARAM) L"Synth Effects");
        SendMessage(hCtrl, dwMsg, 0, (LPARAM) L"Ethnic");
        SendMessage(hCtrl, dwMsg, 0, (LPARAM) L"Percussive");
        SendMessage(hCtrl, dwMsg, 0, (LPARAM) L"Sound effects");
    }
}

// Fill a ComboBox or ListBox with midi channels
void Midi_ListChannels(IN HWND hDlg, IN long CtlId) {

    long c = 0; HWND hCtrl = 0; DWORD dwMsg1 = 0, dwMsg2 = 0;
    WCHAR zTxt[64] = { 0 };

    hCtrl = GetDlgItem(hDlg, CtlId);
    GetClassName(hCtrl, zTxt, sizeof(zTxt)); // see what kind of control to fill
    CharUpper(zTxt);
    if (_wcsicmp(zTxt, L"COMBOBOX") == 0) {         // if to add data to a ComboBox
        SendMessage(hCtrl, CB_RESETCONTENT, 0, 0);
        dwMsg1 = CB_ADDSTRING;
        dwMsg2 = CB_SETCURSEL;
    } else if (_wcsicmp(zTxt, L"LISTBOX") == 0) {  // elseif to add data to a ListBox
        SendMessage(hCtrl, LB_RESETCONTENT, 0, 0);
        dwMsg1 = LB_ADDSTRING;
        dwMsg2 = LB_SETCURSEL;
    }

    for (c = 1; c < 17; c++) {
        wcscpy_s(zTxt, L" Channel"); wcsncat_s(zTxt, STRL(c), _TRUNCATE);
        if (c == 10) {
            wcsncat_s(zTxt, L"  (Percussion set)", _TRUNCATE);
        }
        SendMessage(hCtrl, dwMsg1, 0, (LPARAM) zTxt);
    }
    SendMessage(hCtrl, dwMsg2, 0, 0);
}

// Initialize keyboard memDC. Look worse than it is...  or maybe not. :)
void DrawKeyBoard(IN HWND hWnd) {
    long b = 0, c = 0, d = 0, h = 0, hb = 0, w = 0, wb = 0, nColor1 = 0, nColor2 = 0;
    HDC hDC = 0; HBRUSH hBrush = 0; HPEN hPen = 0, hOldPen = 0; HFONT tmpFont = 0;
    RECT rc;
    POINT pts[8] = { 0 };

    if (g_kbd.memDC) {  // restore and delete what we have created
        if (g_kbd.hBitOld) { SelectObject(g_kbd.memDC, g_kbd.hBitOld); }
        if (g_kbd.hBitmap) { DeleteObject(g_kbd.hBitmap); }
        DeleteDC(g_kbd.memDC);
    }

    hDC = GetDC(hWnd);
    GetClientRect(hWnd, &rc);
    g_kbd.memDC   = CreateCompatibleDC(hDC);

    g_kbd.hBitmap = CreateCompatibleBitmap(hDC, rc.right, rc.bottom);
    g_kbd.hBitOld = (HBITMAP) SelectObject(g_kbd.memDC, g_kbd.hBitmap);
    ReleaseDC(hWnd, hDC);

    FillRect(g_kbd.memDC, &rc, (HBRUSH) GetStockObject(DKGRAY_BRUSH));
    tmpFont = (HFONT) SelectObject(g_kbd.memDC, g_kbd.hFont);

    if (hRgn[0]) {
        for (c = 0; c < UB; c++) {
            if (hRgn[c]) { DeleteObject(hRgn[c]); }
        }
    }

    d = 0;
    for (c = 0; c <= g_kbd.KeyCount; c++) {
         switch (c % 12) {
         case 1:
         case 3:
         case 6:
         case 8:
         case 10:  // black keys
             break;
         default:  // count white keys
             d++;
             break;
         }
         pk[c].nt = c;
    }

    w = rc.right / d;                 // white key width
    if (d * w > rc.right) { w--; }    // ensure last key is fully inside control
    if (w % 2) { w--; }               // need to ensure clean divide by 2

    if (d * w < rc.right) {           // shrink piano control to visible keys width
        SetWindowPos(hWnd, 0, 0, 0, d * w, rc.bottom, SWP_NOMOVE | SWP_NOZORDER);
    }

    wb = (long) (w * 0.7f);           // black key width
    if (wb % 2) { wb--; }             // need to ensure clean divide by 2
    h  = rc.bottom;                   // white key height
    hb = (long) (h * 0.7f);             // black key height

    //--------------------------------------------------------------
    // draw left white key. We use regions for mouse detection.
    //--------------------------------------------------------------
    hBrush  = (HBRUSH) SelectObject(g_kbd.memDC, GetStockObject(WHITE_BRUSH));
    hPen    = CreatePen(PS_SOLID, 1, RGB(96,96,96));
    hOldPen = (HPEN) SelectObject(g_kbd.memDC, hPen);

    pts[0].x =  0;            pts[0].y = 0;   // Region coordinates for Left white key
    pts[1].x =  w - wb / 2;   pts[1].y = 0;
    pts[2].x =  pts[1].x;     pts[2].y = hb;
    pts[3].x =  w;            pts[3].y = pts[2].y;
    pts[4].x =  w;            pts[4].y = h;
    pts[5].x =  0;            pts[5].y = h;

    b = 0; d = 0;
    for (c = 0; c <= g_kbd.KeyCount; c++) {
        if ((c > 0) && (c % 12 == 0)) { b = b + 7 * w; }
        switch (c % 12) {
        case 0:
        case 5:
            switch (c % 12) {
                case 0: d = 0; break;
                case 5: d = 3; break;
            }
            hRgn[c] = CreatePolygonRgn(&pts[0], 6, ALTERNATE);  // create region
            OffsetRgn(hRgn[c], b + d * w, 0);                   // move region into place
            GetRgnBox(hRgn[c], &rc);                            // get region rect

            pk[c].rc = rc;                                      // and use rect for RoundRect draw
            RoundRect(g_kbd.memDC, rc.left, rc.top, rc.right, rc.bottom, 5, 5);
            if (g_kbd.KeyText) { nColor1 = LTGRAY; } else { nColor1 = WHITE; }
            if (g_kbd.NoteText) { nColor2 = GRAY; } else { nColor2 = WHITE; }
            PrintKeyText(pk[c], nColor1, WHITE, nColor2, WHITE);
            break;
        }
    }

    //--------------------------------------------------------------
    // Draw middle white key
    //--------------------------------------------------------------
    pts[0].x =  wb / 2           ; pts[0].y = 0;
    pts[1].x =  w - wb / 2       ; pts[1].y = 0;
    pts[2].x =  pts[1].x         ; pts[2].y = hb;
    pts[3].x =  w                ; pts[3].y = pts[2].y;
    pts[4].x =  w                ; pts[4].y = h;
    pts[5].x =  0                ; pts[5].y = h;
    pts[6].x =  0                ; pts[6].y = hb;
    pts[7].x = pts[0].x          ; pts[7].y = hb;

    b = 0; d = 0;
    for (c = 1; c <= g_kbd.KeyCount; c++) {
          switch (c % 12) {
          case 2:
          case 7:
          case 9:
              switch (c % 12) {
              case 2: d = 1; break;
              case 7: d = 4; break;
              case 9: d = 5; break;
              }
              hRgn[c] = CreatePolygonRgn(&pts[0], 8, ALTERNATE);
              OffsetRgn(hRgn[c], b + d * w, 0);
              GetRgnBox(hRgn[c], &rc);
              pk[c].rc = rc;
              RoundRect(g_kbd.memDC, rc.left, rc.top, rc.right, rc.bottom, 5, 5);
              if (g_kbd.KeyText) { nColor1 = LTGRAY; } else { nColor1 = WHITE; }
              if (g_kbd.NoteText) { nColor2 = GRAY; } else { nColor2 = WHITE; }
              PrintKeyText(pk[c], nColor1, WHITE, nColor2, WHITE);
              break;
          }
          if (c % 12 == 0) { b = b + 7 * w; }
    }

    //------------------------------------------------------------------
    // draw Right white key
    //------------------------------------------------------------------
    pts[0].x =  wb / 2     ; pts[0].y = 0;
    pts[1].x =  w          ; pts[1].y = 0;
    pts[2].x =  w          ; pts[2].y = h;
    pts[3].x =  0          ; pts[3].y = h;
    pts[4].x =  0          ; pts[4].y = hb;
    pts[5].x =  wb / 2     ; pts[5].y = hb;

    b = 0; d = 0;
    for (c = 1; c <= g_kbd.KeyCount; c++) {
         switch (c % 12) {
         case 4:
         case 11:
              switch (c % 12) {
              case  4: d = 2; break;
              case 11: d = 6; break;
              }
              hRgn[c] = CreatePolygonRgn(&pts[0], 6, ALTERNATE);
              OffsetRgn(hRgn[c], b + d * w, 0);
              GetRgnBox(hRgn[c], &rc);
              pk[c].rc = rc;
              RoundRect(g_kbd.memDC, rc.left, rc.top, rc.right, rc.bottom, 5, 5);
              if (g_kbd.KeyText) { nColor1 = LTGRAY; } else { nColor1 = WHITE; }
              if (g_kbd.NoteText) { nColor2 = GRAY; } else { nColor2 = WHITE; }
              PrintKeyText(pk[c], nColor1, WHITE, nColor2, WHITE);
              break;
          }
          if (c % 12 == 0) { b = b + 7 * w; }
    }

    hOldPen = (HPEN) SelectObject(g_kbd.memDC, hOldPen);
    DeleteObject(hPen);
    SelectObject(g_kbd.memDC, hBrush);

    //--------------------------------------------------------------
    // draw black keys
    //--------------------------------------------------------------
    pts[0].x = 0          ; pts[0].y = 0;
    pts[1].x = wb         ; pts[1].y = 0;
    pts[2].x = pts[1].x   ; pts[2].y = hb;
    pts[3].x = 0          ; pts[3].y = hb;

    hBrush  = (HBRUSH) SelectObject(g_kbd.memDC, GetStockObject(BLACK_BRUSH));
    hPen    = CreatePen(PS_SOLID, 1, RGB(0,0,0));
    hOldPen = (HPEN) SelectObject(g_kbd.memDC, hPen);

    b = 0; d = 0;
    for (c = 1; c <= g_kbd.KeyCount; c++) { // 0 is white key, C
          if (c % 12 == 0) { b = b + 7 * w; }
          switch (c % 12) {
          case 1:
          case 3:
          case 6:
          case 8:
          case 10:
              switch (c % 12) {
              case 1:  d = 1; break;
              case 3:  d = 2; break;
              case 6:  d = 4; break;
              case 8:  d = 5; break;
              case 10: d = 6; break;
              }
              hRgn[c] = CreatePolygonRgn(&pts[0], 4, ALTERNATE);
              OffsetRgn(hRgn[c], b + d * w - wb / 2, 0);
              GetRgnBox(hRgn[c], &rc);
              pk[c].rc = rc;
              RoundRect(g_kbd.memDC, rc.left, rc.top, rc.right, rc.bottom, 3, 3);
              break;
          }
    }

    hOldPen = (HPEN) SelectObject(g_kbd.memDC, hOldPen);
    DeleteObject(hPen);
    SelectObject(g_kbd.memDC, hBrush);

    //--------------------------------------------------------------
    // draw thin white rectangle inside black keys
    //--------------------------------------------------------------
    hBrush  = (HBRUSH) SelectObject(g_kbd.memDC, GetStockObject(NULL_BRUSH));
    hPen    = CreatePen(PS_SOLID, 1, RGB(255,255,255));
    hOldPen = (HPEN)SelectObject(g_kbd.memDC, hPen);

    for (c = 1; c <= g_kbd.KeyCount; c++) { // 0 is white key, C
        switch (c % 12) {
        case 1:
        case 3:
        case 6:
        case 8:
        case 10:
            GetRgnBox(hRgn[c], &rc);
            InflateRect(&rc, -2, -2);
            Rectangle(g_kbd.memDC, rc.left, rc.top, rc.right, rc.bottom);
            if (g_kbd.KeyText) { nColor1 = GRAY; } else { nColor1 = BLACK; }
            if (g_kbd.NoteText) { nColor2 = LTGRAY; } else { nColor2 = BLACK; }
            PrintKeyText(pk[c], nColor1, BLACK, nColor2, BLACK);
            break;
        }
    }

    hOldPen = (HPEN) SelectObject(g_kbd.memDC, hOldPen);
    DeleteObject(hPen);
    SelectObject(g_kbd.memDC, hBrush);
    if (tmpFont) { SelectObject(g_kbd.memDC, tmpFont); }
}

LRESULT CALLBACK PianoProc(IN HWND hWnd, IN UINT uMsg, IN WPARAM wParam, IN LPARAM lParam) {
    PAINTSTRUCT ps = { 0 }; POINT pt = { 0 }; RECT rc = { 0 };
    HDC hDC = 0; long c = 0, d = 0, nColor1 = 0, nColor2 = 0;
    static long chPitch, CurKey, OldKey;

    switch (uMsg) {
    case WM_CREATE:
        if (midiOutOpen(&g_kbd.hMidi, -1, 0, 0, 0) != MMSYSERR_NOERROR) {
            MessageBox (GetParent(hWnd), L"midiOutOpen failed!", L"Error message!", MB_OK | MB_ICONERROR);
            return 0;
        }

        // Alias Initkbd
        g_kbd.hParent    = GetParent(hWnd);
        g_kbd.CtlId      = GetDlgCtrlID(hWnd);
        g_kbd.hFont      = MakeFontEx(0, L"Arial", 9, 0, FW_BOLD, 0, 0); //GetStockObject(%ANSI_VAR_FONT)
        g_kbd.KeyText    = 1;
        g_kbd.NoteText   = 1;
        g_kbd.octave     = 4;           // start with octave 3 and 5
        g_kbd.KeyCount   = 28;          // zerobased number of keys to draw (28 = 29 keys)
        g_kbd.volume     = 127;         // max sound volume
        g_kbd.instrument = 0;           // start off with piano
        g_kbd.sustain    = 0;           // start off with no sustain
        g_kbd.vibrato    = 0;           // start off with no vibrato
        g_kbd.balance    = 64;          // set balance to centered
        g_kbd.chStop     = 0x80;        // Note off for channel 1 to 16 / 0x80 to 0x8F
        g_kbd.chPlay     = 0x90;        // Note oon for channel 1 to 16 / 0x90 to 0x9F
        g_kbd.chControl  = 0xB0;        // Control mode change for channel 1 to 16 / 0xB0 to 0xBF
        g_kbd.chInstr    = 0xC0;        // Program change (instrument) for channel 1 to 16 / 0xC0 to 0xCF
        g_kbd.chBend     = 0xE0;        // Pitch wheel range (0 TO 127, +/- two notes) for channel 1 to 16 / 0xE0 to 0xEF
        Midi_SetInstrument(g_kbd.hMidi, g_kbd.chInstr, g_kbd.instrument);  // init instrument
        Midi_SetVolume(g_kbd.hMidi, g_kbd.chControl, g_kbd.volume);        // init volume
        Midi_SetBalance(g_kbd.hMidi, g_kbd.chControl, g_kbd.balance);      // init balance

        DrawKeyBoard(hWnd);

        //--------------------------------------------------------------
        chPitch = 64;          // pitch bend start value (64 is no bend)
        CurKey  = -1;
        break;

    case WM_DESTROY: // exit, time to clean up memory, etc
        if (g_kbd.hMidi ) { midiOutClose(g_kbd.hMidi); } // Close the midi device
        if (g_kbd.memDC) {  // restore and delete what we have created
            if (g_kbd.hBitOld) { SelectObject(g_kbd.memDC, g_kbd.hBitOld); }
            if (g_kbd.hBitmap) { DeleteObject(g_kbd.hBitmap); }
            DeleteDC(g_kbd.memDC);
        }
        if (g_kbd.hFont) { DeleteObject(g_kbd.hFont); }
        for (c = 0; c < UB; c++) {
            if (hRgn[c]) { DeleteObject(hRgn[c]); }
        }
        return 0;

    case WM_GETDLGCODE: // Ensure the control processes all keys by itself
        return DLGC_WANTALLKEYS;

    case WM_PAINT:  // on repaint, simply copy g_kbd.memDC to control DC
        BeginPaint(hWnd, &ps);
          GetClientRect(hWnd, &rc);
          BitBlt(ps.hdc, 0, 0, rc.right, rc.bottom, g_kbd.memDC, 0, 0, SRCCOPY);
        EndPaint(hWnd, &ps);
        return 0;

    case WM_KEYDOWN:
        switch (wParam) {
        case VK_ESCAPE: // enable Exit on Esc key
            SendMessage(g_kbd.hParent, WM_COMMAND, MAKEWPARAM(IDCANCEL, BN_CLICKED), 0);
            break;

        case VK_UP:  // Arrow up ^  - Pitch bend up
            if (chPitch == 64) {  // send notification on each change to parent's control id's WM_COMMAND
                for (chPitch = 64; chPitch <= UB; chPitch++) {
                     Midi_PitchBend(g_kbd.hMidi, g_kbd.chBend, chPitch);
                     SendMessage(g_kbd.hParent, WM_COMMAND, MAKEWPARAM(g_kbd.CtlId, PN_PITCHBEND), -(chPitch - 64));
                     //if ((GetAsyncKeyState(VK_UP) && 0x8000) == 0) { break; } // if key is released
                     if (!IsKeyDown(VK_UP)) { break; } // if key is released
                     Sleep(8);
                }
            }
            break;

        case VK_DOWN:  // Arrow down v  - Pitch bend down
            if (chPitch == 64) {
                for (chPitch = 64; chPitch --> 0;) {
                    Midi_PitchBend(g_kbd.hMidi, g_kbd.chBend, chPitch);
                    SendMessage(g_kbd.hParent, WM_COMMAND, MAKEWPARAM(g_kbd.CtlId, PN_PITCHBEND), -(chPitch - 64));
                    //if ((GetAsyncKeyState(VK_DOWN) && 0x8000) == 0) { break; }
                    if (!IsKeyDown(VK_DOWN)) { break; } // if key is released
                    Sleep(8);
                }
            }
            break;

        case VK_LEFT:  // Arrow left < - octave change down
            g_kbd.octave = max(0, g_kbd.octave - 1);
            SendMessage(hWnd, PM_SETOCTAVE, g_kbd.octave, 1);
            SendMessage(g_kbd.hParent, WM_COMMAND, MAKEWPARAM(g_kbd.CtlId, PN_OCTAVE), g_kbd.octave);
            break;

        case VK_RIGHT:  // Arrow right > - octave change up
            g_kbd.octave = min(10 - g_kbd.KeyCount / 12, g_kbd.octave + 1);
            SendMessage(hWnd, PM_SETOCTAVE, g_kbd.octave, 1);
            SendMessage(g_kbd.hParent, WM_COMMAND, MAKEWPARAM(g_kbd.CtlId, PN_OCTAVE), g_kbd.octave);
            break;

        case  VK_0:
        case  VK_1:
        case  VK_2:
        case  VK_3:
        case  VK_4:
        case  VK_5:
        case  VK_6:
        case  VK_7:
        case  VK_8:
        case  VK_9:
        case  VK_A:
        case  VK_B:
        case  VK_C:
        case  VK_D:
        case  VK_E:
        case  VK_F:
        case  VK_G:
        case  VK_H:
        case  VK_I:
        case  VK_J:
        case  VK_K:
        case  VK_L:
        case  VK_M:
        case  VK_N:
        case  VK_O:
        case  VK_P:
        case  VK_Q:
        case  VK_R:
        case  VK_S:
        case  VK_T:
        case  VK_U:
        case  VK_V:
        case  VK_W:
        case  VK_X:
        case  VK_Y:
        case  VK_Z: // See if a note key has been pressed
            c = INSTRPOS(&KEY1[0], CHR$((long) wParam));  // $KEY1 is used for pk array compare
            if (c) {
                c--;
                if (pk[c].dn == 0) {          // if not pressed before
                    g_kbd.note = g_kbd.octave * 12 + pk[c].nt;   // actual note
                    pk[c].dn = 2;                                    // 2 = keyboard key pressed it
                    SendMessage(g_kbd.hParent, WM_COMMAND, MAKEWPARAM(g_kbd.CtlId, PN_PLAYNOTE), g_kbd.note);
                    Midi_PlayNote(g_kbd.hMidi, g_kbd.chPlay, g_kbd.note, g_kbd.volume);
                    switch (c % 12) {
                    case 1:
                    case 3:
                    case 6:
                    case 8:
                    case 10: // black keys
                        if (g_kbd.KeyText) { nColor1 = GRAY; } else { nColor1 = BLACK; }
                        if (g_kbd.NoteText) { nColor2 = YELLOW; } else { nColor2 = BLACK; }
                        PrintKeyText(pk[c], nColor1, BLACK, nColor2, BLACK);
                        break;
                    default: // white keys
                        if (g_kbd.KeyText) { nColor1 = LTGRAY; } else { nColor1 = WHITE; }
                        if (g_kbd.NoteText) { nColor2 = BLUE; } else { nColor2 = WHITE; }
                        PrintKeyText(pk[c], nColor1, WHITE, nColor2, WHITE);
                        break;
                    }
                    InvalidateRect(hWnd, NULL, 0);
                    UpdateWindow(hWnd);
                }
            }
            break;
        }
        return 0;

    case WM_KEYUP:
        switch (wParam) {
        case VK_UP:    // Arrow up - Pitch has been bended up
            if (chPitch > 64) {
                chPitch = 64  ;
                Midi_PitchBend(g_kbd.hMidi, g_kbd.chBend, chPitch);
                SendMessage(g_kbd.hParent, WM_COMMAND, MAKEWPARAM(g_kbd.CtlId, PN_PITCHBEND), 0);
            }
            break;

        case VK_DOWN:  // Arrow down - Pitch has been bended down
            if (chPitch < 64) {
                chPitch = 64;
                Midi_PitchBend(g_kbd.hMidi, g_kbd.chBend, chPitch);
                SendMessage(g_kbd.hParent, WM_COMMAND, MAKEWPARAM(g_kbd.CtlId, PN_PITCHBEND), 0);
            }
            break;

        case  VK_0:
        case  VK_1:
        case  VK_2:
        case  VK_3:
        case  VK_4:
        case  VK_5:
        case  VK_6:
        case  VK_7:
        case  VK_8:
        case  VK_9:
        case  VK_A:
        case  VK_B:
        case  VK_C:
        case  VK_D:
        case  VK_E:
        case  VK_F:
        case  VK_G:
        case  VK_H:
        case  VK_I:
        case  VK_J:
        case  VK_K:
        case  VK_L:
        case  VK_M:
        case  VK_N:
        case  VK_O:
        case  VK_P:
        case  VK_Q:
        case  VK_R:
        case  VK_S:
        case  VK_T:
        case  VK_U:
        case  VK_V:
        case  VK_W:
        case  VK_X:
        case  VK_Y:
        case  VK_Z: // same as on key down, but here we stop playing note and reset key text instead.
            c = INSTRPOS(&KEY1[0], CHR$((long) wParam));  // see if it is a defined key
            if (c) {
                c--;
                if (pk[c].dn == 2) {                // if it was pressed by a keyboard key
                    SendMessage(g_kbd.hParent, WM_COMMAND, MAKEWPARAM(g_kbd.CtlId, PN_STOPNOTE), g_kbd.note);
                    Midi_StopNote(g_kbd.hMidi, g_kbd.chStop, g_kbd.octave * 12 + pk[c].nt);
                    pk[c].dn = 0;

                    switch (c % 12) { // reset key text
                    case 1:
                    case 3:
                    case 6:
                    case 8:
                    case 10:  // black keys
                        if (g_kbd.KeyText) { nColor1 = GRAY; } else { nColor1 = BLACK; }
                        if (g_kbd.NoteText) { nColor2 = LTGRAY; } else { nColor2 = BLACK; }
                        PrintKeyText(pk[c], nColor1, BLACK, nColor2, BLACK);
                        break;
                    default:  // white keys
                        if (g_kbd.KeyText) { nColor1 = LTGRAY; } else { nColor1 = WHITE; }
                        if (g_kbd.NoteText) { nColor2 = GRAY; } else { nColor2 = WHITE; }
                        PrintKeyText(pk[c], nColor1, WHITE, nColor2, WHITE);
                        break;
                    }
                    InvalidateRect(hWnd, NULL, 0);
                    UpdateWindow(hWnd);
                    return 0;
                }
            }
            break;
        }
        break;

    case WM_LBUTTONDOWN:
    case WM_LBUTTONDBLCLK:
        SetCapture(hWnd);
        pt.x = LOINT(lParam);
        pt.y = HIINT(lParam);
        for (c = 0; c < UB; c++) {
            if ((c != CurKey) && (PtInRegion(hRgn[c], pt.x, pt.y))) {
                pk[c].dn = 1; // 1 = mouse key pressed it
                CurKey = c;
                g_kbd.note = g_kbd.octave * 12 + CurKey;
                SendMessage(g_kbd.hParent, WM_COMMAND, MAKEWPARAM(g_kbd.CtlId, PN_PLAYNOTE), g_kbd.note);
                Midi_PlayNote(g_kbd.hMidi, g_kbd.chPlay, g_kbd.note, g_kbd.volume);
                OldKey = CurKey;

                switch (CurKey % 12) {
                case 1:
                case 3:
                case 6:
                case 8:
                case 10: // black keys
                    if (g_kbd.KeyText) { nColor1 = GRAY; } else { nColor1 = BLACK; }
                    if (g_kbd.NoteText) { nColor2 = YELLOW; } else { nColor2 = BLACK; }
                    PrintKeyText(pk[c], nColor1, BLACK, nColor2, BLACK);
                    break;
                default: // white keys
                    if (g_kbd.KeyText) { nColor1 = LTGRAY; } else { nColor1 = WHITE; }
                    if (g_kbd.NoteText) { nColor2 = RED; } else { nColor2 = WHITE; }
                    PrintKeyText(pk[c], nColor1, WHITE, nColor2, WHITE);
                    break;
                }
                InvalidateRect(hWnd, NULL, 0);
                UpdateWindow(hWnd);
                break;
            }
        }
        break;

    case WM_MOUSEMOVE:
        if (IsLButtonDown()) {
        //if ((GetAsyncKeyState(VK_LBUTTON) && (0x8000))) {
            if (GetCapture() != hWnd) { SetCapture(hWnd); }  // ensure mouse capture

            pt.x = LOINT(lParam);
            pt.y = HIINT(lParam);
            for (c = 0; c < UB; c++) {
                if ((c != CurKey) && (PtInRegion(hRgn[c], pt.x, pt.y))) {
                    CurKey = c;
                    break;
                }
            }
            if (CurKey != OldKey) {
                pk[OldKey].dn = 0; // 1 = mouse key pressed it
                g_kbd.note = g_kbd.octave * 12 + OldKey;
                Midi_StopNote(g_kbd.hMidi, g_kbd.chStop, g_kbd.note); // stop playing this note

                switch (OldKey % 12) { // reset key text
                case 1:
                case 3:
                case 6:
                case 8:
                case 10: // black keys
                    if (g_kbd.KeyText) { nColor1 = GRAY; } else { nColor1 = BLACK; }
                    if (g_kbd.NoteText) { nColor2 = LTGRAY; } else { nColor2 = BLACK; }
if (OldKey < 0) { OldKey = CurKey; }

                    PrintKeyText(pk[OldKey], nColor1, BLACK, nColor2, BLACK);
                    break;
                default: // white keys
                    if (g_kbd.KeyText) { nColor1 = LTGRAY; } else { nColor1 = WHITE; }
                    if (g_kbd.NoteText) { nColor2 = GRAY; } else { nColor2 = WHITE; }
if (OldKey < 0) { OldKey = CurKey; }
                    PrintKeyText(pk[OldKey], nColor1, WHITE, nColor2, WHITE);
                    break;
                }

                pk[CurKey].dn = 1; // 1 = mouse key pressed it
                g_kbd.note = g_kbd.octave * 12 + CurKey;
                Midi_PlayNote(g_kbd.hMidi, g_kbd.chPlay, g_kbd.note, g_kbd.volume);
                switch (CurKey % 12) {
                case 1:
                case 3:
                case 6:
                case 8:
                case 10: // black keys
                    if (g_kbd.KeyText) { nColor1 = GRAY; } else { nColor1 = BLACK; }
                    if (g_kbd.NoteText) { nColor2 = YELLOW; } else { nColor2 = BLACK; }
                    PrintKeyText(pk[CurKey], nColor1, BLACK, nColor2, BLACK);
                    break;
                default: // white keys
                    if (g_kbd.KeyText) { nColor1 = LTGRAY; } else { nColor1 = WHITE; }
                    if (g_kbd.NoteText) { nColor2 = RED; } else { nColor2 = WHITE; }
                    PrintKeyText(pk[CurKey], nColor1, WHITE, nColor2, WHITE);
                    break;
                }
                InvalidateRect(hWnd, NULL, 0);
                UpdateWindow(hWnd);
                OldKey = CurKey;
            }
        }
        break;

    case WM_LBUTTONUP:
        pk[CurKey].dn = 0;
        g_kbd.note = g_kbd.octave * 12 + CurKey;
        SendMessage(g_kbd.hParent, WM_COMMAND, MAKEWPARAM(g_kbd.CtlId, PN_STOPNOTE), g_kbd.note);
        Midi_StopNote(g_kbd.hMidi, g_kbd.chStop, g_kbd.note); // and stop playing this note

        switch (CurKey % 12) { // reset key text
        case 1:
        case 3:
        case 6:
        case 8:
        case 10: // black keys
            if (g_kbd.KeyText) { nColor1 = GRAY; } else { nColor1 = BLACK; }
            if (g_kbd.NoteText) { nColor2 = LTGRAY; } else { nColor2 = BLACK; }
            PrintKeyText(pk[CurKey], nColor1, BLACK, nColor2, BLACK);
            break;
        default:  // white keys
            if (g_kbd.KeyText) { nColor1 = LTGRAY; } else { nColor1 = WHITE; }
            if (g_kbd.NoteText) { nColor2 = GRAY; } else { nColor2 = WHITE; }
            PrintKeyText(pk[CurKey], nColor1, WHITE, nColor2, WHITE);
            break;
        }

        InvalidateRect(hWnd, NULL, 0);
        UpdateWindow(hWnd);
        CurKey = -1;
        OldKey = -1;
        ReleaseCapture();
        break;

    //------------------------------------------------------------------
    // a bunch of useful PM_ Piano Messages
    //------------------------------------------------------------------
    case PM_SETDEVICE:
        // can't test this one, because like most, I only have one midi device. Hope it works..
        if (g_kbd.hMidi) { midiOutClose(g_kbd.hMidi); } // Close current midi device
        if (midiOutOpen((LPHMIDIOUT) g_kbd.hMidi, (UINT) wParam, 0, 0, 0) != MMSYSERR_NOERROR) {
            MessageBox(GetParent(hWnd), L"midiOutOpen failed - trying default device instead!", L"Error message!", MB_ICONERROR);
            midiOutOpen((LPHMIDIOUT) g_kbd.hMidi, -1, 0, 0, 0);
        }
        Midi_SetInstrument(g_kbd.hMidi, g_kbd.chInstr, g_kbd.instrument);  // reset selected instrument
        Midi_SetVolume(g_kbd.hMidi, g_kbd.chControl, g_kbd.volume);        // reset volume
        Midi_SetBalance(g_kbd.hMidi, g_kbd.chControl, g_kbd.balance);      // reset balance
        Midi_SetSustain(g_kbd.hMidi, g_kbd.chControl, g_kbd.sustain);      // reset sustain
        SetFocus(hWnd);
        return 0;

    case PM_SETCHANNEL:
        Midi_StopAllNotes(g_kbd.hMidi, g_kbd.chControl);  // stop all notes
        Midi_SetSustain(g_kbd.hMidi, g_kbd.chControl, 0); // reset sustain

        g_kbd.chControl = 0xB0 + (long) wParam;
        g_kbd.chInstr   = 0xC0 + (long) wParam;
        g_kbd.chStop    = 0x80 + (long) wParam;
        g_kbd.chPlay    = 0x90 + (long) wParam;
        g_kbd.chBend    = 0xE0 + (long) wParam;

        Midi_SetInstrument(g_kbd.hMidi, g_kbd.chInstr, g_kbd.instrument);  // reset selected instrument
        Midi_SetVolume(g_kbd.hMidi, g_kbd.chControl, g_kbd.volume);        // reset volume
        Midi_SetBalance(g_kbd.hMidi, g_kbd.chControl, g_kbd.balance);      // reset balance
        Midi_SetSustain(g_kbd.hMidi, g_kbd.chControl, g_kbd.sustain);      // reset sustain
        SetFocus(hWnd);
        return 0;

    case PM_LISTDEVICES:     // wParam is list id
        Midi_ListDevices(g_kbd.hParent, (long) wParam);
        return 0;

    case PM_LISTCHANNELS:    // wParam is list id
        Midi_ListChannels(g_kbd.hParent, (long) wParam);
        return 0;

    case PM_LISTPATCHES:     // wParam is list id
        Midi_ListPatches(g_kbd.hParent, (long) wParam);
        return 0;

    case PM_LISTINSTRUMENTS: // wParam is list id, lParam is patches number
        Midi_ListInstruments(g_kbd.hParent, (long) wParam, (long) lParam);
        return 0;

    case PM_GETOCTAVE:  //returns currently selected base octave
        return g_kbd.octave;

    case PM_SETOCTAVE:  // sets 0 to 9 octaves
        Midi_StopAllNotes(g_kbd.hMidi, g_kbd.chControl);   // stop all notes
        g_kbd.octave = (long) wParam;
        for (c = 0; c <UB; c++) {
            if (pk[c].dn) {
                g_kbd.note = g_kbd.octave * 12 + pk[c].nt; // refresh note
                Midi_PlayNote(g_kbd.hMidi, g_kbd.chPlay, g_kbd.note, g_kbd.volume);
            }
        }
        if (lParam) { SetFocus(hWnd); }
        return 0;

    case PM_SETINSTRUMENT:  // 0 to 127 instruments
        for (c = 0; c < UB; c++) { // On change, stop all notes
            Midi_StopNote(g_kbd.hMidi, g_kbd.chStop, g_kbd.octave * 12 + pk[c].nt);
            pk[c].dn = 0;
        }
        g_kbd.instrument = (long) wParam;
        Midi_SetInstrument(g_kbd.hMidi, g_kbd.chInstr, g_kbd.instrument);  //set selected instrument
        if (lParam) { SetFocus(hWnd); }
        return 0;

    case PM_SETPITCH:   // -64 to 64, where 0 is normal pitch
        Midi_PitchBend(g_kbd.hMidi, g_kbd.chBend, (long) wParam);
        if (lParam) { SetFocus(hWnd); }
        return 0;

    case PM_SETVOLUME:  // 0 to 127, where 127 is max sound volume
        g_kbd.volume = (long) wParam;
        Midi_SetVolume(g_kbd.hMidi, g_kbd.chControl, g_kbd.volume);
        if (lParam) { SetFocus(hWnd); }
        return 0;

    case PM_SETBALANCE:  // -64 to 64, where 0 is centered
        g_kbd.balance = (long) wParam;
        Midi_SetBalance(g_kbd.hMidi, g_kbd.chControl, g_kbd.balance);
        if (lParam) { SetFocus(hWnd); }
        return 0;

    case PM_SETVIBRATO:  // 0 to 127, where 0 is no vibrato
        g_kbd.vibrato = (long) wParam;
        Midi_SetModulation(g_kbd.hMidi, g_kbd.chControl, g_kbd.vibrato);
        if (lParam) { SetFocus(hWnd); }
        return 0;

    case PM_SETSUSTAIN:  // on/off, where anything below 64 is off (0/127 = off/on)
        g_kbd.sustain = (long) wParam;
        Midi_SetSustain(g_kbd.hMidi, g_kbd.chControl, g_kbd.sustain);
        if (lParam) { SetFocus(hWnd); }
        return 0;

    case PM_GETKEYCOUNT:
        return g_kbd.KeyCount + 1;

    case PM_SETKEYCOUNT:  // set total number of keys to draw/play
        g_kbd.KeyCount = min(127, (long) wParam - 1);
        DrawKeyBoard(hWnd);
        InvalidateRect(hWnd, NULL, 0);
        UpdateWindow(hWnd);
        if (lParam) { SetFocus(hWnd); }
        return 0;

    case PM_SETKEYTEXT:  // if to print keyboard key letter
        g_kbd.KeyText = (long) wParam;
        DrawKeyBoard(hWnd);
        InvalidateRect(hWnd, NULL, 0);
        UpdateWindow(hWnd);
        if (lParam) { SetFocus(hWnd); }
        return 0;

    case PM_SETNOTETEXT:  // if to print key note
        g_kbd.NoteText = (long) wParam;
        DrawKeyBoard(hWnd);
        InvalidateRect(hWnd, NULL, 0);
        UpdateWindow(hWnd);
        if (lParam) { SetFocus(hWnd); }
        return 0;

    case PM_PLAYNOTE:  // to enable playing notes via this message
        g_kbd.note   = min(127, (long) wParam); // play this note
        g_kbd.volume = min(127, (long) lParam); // 127 is max volume
        Midi_PlayNote(g_kbd.hMidi, g_kbd.chPlay, g_kbd.note, g_kbd.volume);
        break;

    case PM_STOPNOTE:  // to enable stop playing notes via this message
        g_kbd.note = min(127, (long) wParam);
        Midi_StopNote(g_kbd.hMidi, g_kbd.chStop, g_kbd.note); // stop playing this note
        break;

    }
    return DefWindowProc(hWnd, uMsg, wParam, lParam);
}

//====================================================================
// Initialize and create the control in one and same call
//--------------------------------------------------------------------
HWND CreatePiano(IN HWND hParent, IN long CtrlId, IN WCHAR* zTxt, IN long cLeft, IN long cTop,
                 IN long cWidth, IN long cHeight, IN DWORD dwStyle, IN DWORD dwStyleEx) {
    HWND hCtrl = 0;
    WCHAR zClass[] = L"PianoCtrl";
    WNDCLASSEX wcx = { 0 };
    HINSTANCE hInstance = (HINSTANCE) GetWindowLongPtr(hParent, GWL_HINSTANCE);
    wcx.cbSize = sizeof(wcx);
    long IsInitialized = GetClassInfoEx(hInstance, zClass, &wcx);
    if (!IsInitialized) {
        wcx.style = CS_DBLCLKS;
        wcx.lpfnWndProc = PianoProc;
        wcx.cbClsExtra = 0;
        wcx.cbWndExtra = 4;
        wcx.hInstance = hInstance;
        wcx.hIcon = LoadIcon(NULL, IDI_APPLICATION);
        wcx.hCursor = LoadCursor(NULL, IDC_ARROW);
        wcx.hbrBackground = (HBRUSH) COLOR_BTNSHADOW;
        wcx.lpszMenuName = NULL;
        wcx.lpszClassName = zClass;
        wcx.hIconSm = wcx.hIcon;
        if (RegisterClassEx(&wcx)) { IsInitialized = -1; }
    }
    if (IsInitialized) {
        hCtrl = CreateWindowEx(dwStyleEx, zClass, zTxt, dwStyle, cLeft, cTop, cWidth, cHeight, hParent, (HMENU) CtrlId, wcx.hInstance, NULL);
    }
    return hCtrl;
}
